<table>
    <tr>
        <td>
            <table>
                <thead>
                    <tr>
                        <td> </td>
                        <th> Array </th>
                        <th> Chunk </th>
                    </tr>
                </thead>
                <tbody>
                    {% if nbytes %}
                    <tr>
                        <th> Bytes </th>
                        <td> {{ nbytes }} </td>
                        <td> {{ cbytes }} </td>
                    </tr>
                    {% endif %}
                    <tr>
                        <th> Shape </th>
                        <td> {{ array.shape }} </td>
                        <td> {{ array.chunksize }} </td>
                    </tr>
                    <tr>
                        <th> Count </th>
                        <td> {{ arrs_in_plan }} </td>
                        <td> {{ array.npartitions }} Chunks </td>
                    </tr>
                    <tr>
                        <th> Type </th>
                        <td> {{ array.dtype }} </td>
                        <td> {{ arrtype }} </td>
                    </tr>
                </tbody>
            </table>
        </td>
        <td>
        {{grid}}
        </td>
    </tr>
</table>
